from typing import Any, Dict, List, Callable

from .api import API, FileToken


class ResourceNotSynchronized(RuntimeError):
    pass


class Resource:
    def __init__(self, api: API, initial_json: Dict[str, Any]):
        self._api = api
        self._json = initial_json
        self._json_new = {}

    # Abstract methods to override
    def _get_instance_urlparts(self) -> List[str]:
        """URL Parts generator for this specific instance"""
        raise NotImplementedError()

    @classmethod
    def _get_class_urlparts(cls) -> List[str]:
        """URL Parts Generator for this class"""
        raise NotImplementedError()

    def _setter_transforms(self) -> Dict[str, Callable]:
        """Converts set value to JSON-serializable dictionary"""
        return {}

    def _getter_transforms(self) -> Dict[str, Callable]:
        """Converts internal JSON dictionary into usable value"""
        return {}

    def _serialized(self) -> Dict[str, Any]:
        """Return formatted JSON dictionary for PUT/POST request body"""
        return {}

    # Common methods
    def _copy_new_json(self, keys_to_copy: List[str]):
        ret = {}
        for key in keys_to_copy:
            if key in self._json_new:
                ret[key] = self._json_new[key]
        return ret

    def _update_json(self, data: Dict[str, Any], force: bool = False):
        if not force:
            for key in data:
                if key not in self._json_new:
                    # property not changed
                    continue
                if self._json[key] == self._json_new[key]:
                    # property set back to original
                    continue
                if data[key] == self._json_new[key]:
                    # property set to new value
                    continue
                raise ResourceNotSynchronized(key)
        self._json_new = {}
        self._json = data

    def pull(self):
        data = self._api._call("GET", self._get_instance_urlparts())
        self._update_json(data)

    def push(self):
        body = self._serialized()
        if "version" in self._json and self._json["version"]:
            body["version"] = self._json["version"]
            data = self._api._call("PUT", self._get_instance_urlparts(), body=body)
        else:
            data = self._api._call("POST", self._get_class_urlparts(), body=body)
        self._update_json(data, force=True)

    @classmethod
    def _apply_transforms(cls, transforms: Dict[str, Callable], property_name: str, property_value: Any) -> Any:
        if property_value is None:
            return None
        elif isinstance(property_value, list):
            return [cls._apply_transforms(transforms, property_name, x) for x in property_value]
        elif property_name in transforms:
            return transforms[property_name](property_value)
        else:
            return property_value

    def _generic_getter(self, property_name: str, dynamic_refresh: bool = True) -> Any:
        if property_name in self._json_new:
            return self._apply_transforms(
                self._getter_transforms(), property_name, self._json_new[property_name]
            )
        elif property_name in self._json:
            return self._apply_transforms(self._getter_transforms(), property_name, self._json[property_name])
        elif dynamic_refresh:
            self.pull()
            return self._generic_getter(property_name, False)
        else:
            raise KeyError(f"{property_name} is not present in the JSON response")

    def _generic_setter(self, property_name: str, property_value: Any, dynamic_refresh: bool = True) -> None:
        if property_name in self._json:
            if isinstance(self._json[property_name], list):
                if not isinstance(property_value, list):
                    raise ValueError(f"{property_name} must be a list, not {type(property_value)}")
            self._json_new[property_name] = self._apply_transforms(
                self._setter_transforms(), property_name, property_value
            )
        elif dynamic_refresh:
            self.pull()
            self._generic_setter(property_name, property_value, False)
        else:
            raise KeyError(f"{property_name} is not present in the JSON response")

    def _file_getter(self, property_name: str, dynamic_refresh: bool = True) -> str:
        if f"{property_name}Url" in self._json:
            return self._api._create_data_url(self._json[f"{property_name}Url"])
        elif dynamic_refresh:
            self.pull()
            return self._file_getter(property_name, False)
        else:
            raise KeyError(f"{property_name} is not a URL resource in the JSON response")

    def _file_setter(
        self, property_name: str, property_value: FileToken, dynamic_refresh: bool = False
    ) -> None:
        self._json_new[f"{property_name}Token"] = property_value.token



from __future__ import annotations
from typing import Any, Dict, List, Callable

from .api import API
from .resource import Resource, ResourceNotSynchronized


class Tag(Resource):
    # Implementing Abstract Methods
    def _get_instance_urlparts(self) -> List[str]:
        return ["tag", self._json["names"][0]]

    @classmethod
    def _get_class_urlparts(cls) -> List[str]:
        return ["tags"]

    def _setter_transforms(self) -> Dict[str, Callable]:
        return {
            "implications": lambda x: {"names": x.names, "category": x.category},
            "suggestions": lambda x: {"names": x.names, "category": x.category},
        }

    def _getter_transforms(self) -> Dict[str, Callable]:
        return {
            "implications": lambda x: Tag(self._api, x),
            "suggestions": lambda x: Tag(self._api, x),
        }

    def _serialized(self) -> Dict[str, Any]:
        ret = self._copy_new_json(["names", "category", "description", "implications", "suggestions"])
        if "implications" in ret:
            ret["implications"] = [x.primary_name for x in ret["implications"]]
        if "suggestions" in ret:
            ret["suggestions"] = [x.primary_name for x in ret["suggestions"]]
        return ret

    # Factory Methods
    @classmethod
    def from_name(cls, api: API, name: str) -> Tag:
        t = cls(api, {"names": [name]})
        t.pull()
        return t

    @classmethod
    def new(cls, api: API, name: str) -> Tag:
        # Get default tag category
        tag_cats = [x["name"] for x in api._call("GET", ["tag-categories"])["results"] if x["default"]]
        assert len(tag_cats) == 1
        default_cat = tag_cats[0]
        # Create and return tag
        t = cls(api, {})
        t._json_new = {"names": [name], "category": default_cat}
        t.push()
        return t

    def merge_from(self, source: Tag, add_as_alias: bool) -> None:
        """
        Merges source tag into this tag
        """
        if "version" not in source._json or source._json_new:
            raise ResourceNotSynchronized("Target tag is not synchronized")
        if "version" not in self._json or self._json_new:
            raise ResourceNotSynchronized("This tag is not synchronized")
        data = self._api._call(
            "POST",
            ["tag-merge"],
            body={
                "removeVersion": source._json["version"],
                "remove": source.primary_name,
                "mergeToVersion": self._json["version"],
                "mergeTo": self.primary_name,
            },
        )
        self._update_json(data, force=True)
        if add_as_alias:
            n = self.names
            n.extend([x for x in source._json["names"] if x not in n])
            self.names = n
            self.push()
        source._json = {}

    # Getters and Setters
    @property
    def names(self) -> List[str]:
        return self._generic_getter("names")

    @names.setter
    def names(self, val: List[str]) -> None:
        self._generic_setter("names", val)

    @property
    def category(self) -> str:
        return self._generic_getter("category")

    @category.setter
    def category(self, val: str) -> None:
        self._generic_setter("category", val)

    @property
    def implications(self) -> List[Tag]:
        return self._generic_getter("implications")

    @implications.setter
    def implications(self, val: List[Tag]) -> None:
        self._generic_setter("implications", val)

    @property
    def suggestions(self) -> List[Tag]:
        return self._generic_getter("suggestions")

    @suggestions.setter
    def suggestions(self, val: List[Tag]) -> None:
        self._generic_setter("suggestions", val)

    @property
    def description(self) -> str:
        return self._generic_getter("description")

    @description.setter
    def description(self, val: str) -> None:
        self._generic_setter("description", val)

    @property
    def primary_name(self) -> str:
        return self.names[0]

    @primary_name.setter
    def primary_name(self, val: str) -> None:
        existing_names = self.names
        if val in existing_names:
            existing_names.remove(val)
        existing_names.insert(0, val)
        self.names = existing_names

    def add_name(self, val: str) -> None:
        existing_names = self.names
        if val not in existing_names:
            existing_names.append(val)
        self.names = existing_names



from typing import List, Generator

from collections import namedtuple
from contextlib import nullcontext

from tqdm import tqdm

from .api import API, FileToken
from .resource import Resource
from .tag import Tag
from .post import Post


SearchResult = namedtuple("SearchResult", ["post", "distance", "exact"], defaults=[False])


def _search_generic(
    api: API,
    search_query: str,
    transforming_class: type,
    page_size: int,
    show_progress_bar: bool = False,
) -> Generator[Resource, None, None]:
    offset = 0
    total = None
    with (tqdm() if show_progress_bar else nullcontext()) as pbar:
        while True:
            page = api._call(
                "GET",
                transforming_class._get_class_urlparts(),
                urlquery={"offset": offset, "limit": page_size, "query": search_query},
            )
            offset = offset + len(page["results"])
            if page["total"] != total:
                total = page["total"]
                if show_progress_bar:
                    pbar.total = total
                    pbar.refresh()
            for item in page["results"]:
                if show_progress_bar:
                    pbar.update()
                yield transforming_class(api, item)
            if offset >= total:
                break


def search_tag(
    api: API, search_query: str, page_size: int = 20, show_progress_bar: bool = False
) -> Generator[Tag, None, None]:
    return _search_generic(api, search_query, Tag, page_size, show_progress_bar)


def search_post(
    api: API, search_query: str, page_size: int = 20, show_progress_bar: bool = False
) -> Generator[Post, None, None]:
    return _search_generic(api, search_query, Post, page_size, show_progress_bar)


def search_by_image(api: API, image: FileToken) -> List[SearchResult]:
    result = api._call("POST", ["posts", "reverse-search"], body={"contentToken": image.token})
    ret = [SearchResult(post=Post(api, x["post"]), distance=x["distance"]) for x in result["similarPosts"]]
    if result["exactPost"]:
        ret.insert(0, SearchResult(post=Post(api, result["exactPost"]), distance=None, exact=True))
    return ret



from __future__ import annotations
from typing import Any, Dict, List

from base64 import b64encode
import json
import os
import re
from urllib.parse import urlencode, urlsplit, urlunsplit, urljoin, quote

from appdirs import user_data_dir
import requests


class FileToken:
    def __init__(self, token: str, filepath: str):
        self._token = token
        self._filepath = filepath

    @property
    def token(self):
        return self._token

    @property
    def filepath(self):
        return self._filepath

    def __str__(self) -> str:
        return f"<Upload token for file at {self._filepath}>"


class SzurubooruHTTPError(requests.exceptions.HTTPError):
    pass


class API:
    _token_checker = re.compile(r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$")

    @staticmethod
    def _encode_auth_headers(u: str, p: str) -> str:
        return b64encode(f"{u}:{p}".encode("utf-8")).decode("ascii")

    @staticmethod
    def _check_api_response(r: requests.models.Response) -> None:
        if not r.status_code == requests.codes.ok:
            try:
                msg = r.json()
                msg = f"{msg['name']}: {msg['description']}"
            except ValueError:
                msg = r.text
            raise SzurubooruHTTPError(msg)

    def __init__(
        self,
        base_url: str,
        username: str = None,
        password: str = None,
        token: str = None,
        api_url: str = "api",
    ):
        # Extract Base URL parts
        parsed_base_url = urlsplit(base_url)
        self._url_scheme = parsed_base_url.scheme
        self._url_netloc = parsed_base_url.netloc
        if not (self._url_scheme and self._url_netloc):
            raise ValueError("Base URL is not valid")
        if self._url_scheme not in ("http", "https"):
            raise ValueError("Base URL must be of HTTP or HTTPS scheme")
        self._url_path_prefix = parsed_base_url.path.rstrip("/")

        # Extract API URL parts
        parsed_api_url = urlsplit(api_url)
        self._api_scheme = parsed_api_url.scheme or self._url_scheme
        self._api_netloc = parsed_api_url.netloc or self._url_netloc
        if self._api_scheme not in ("http", "https"):
            raise ValueError("API URL must be of HTTP or HTTPS scheme")
        if parsed_api_url.path.startswith("/") or self._api_netloc != self._url_netloc:
            self._api_path_prefix = parsed_api_url.path.rstrip("/")
        else:
            self._api_path_prefix = f"{self._url_path_prefix}/{parsed_api_url.path}".rstrip("/")

        # Extract Auth Info
        self._api_headers = {"Accept": "application/json"}
        username = username or parsed_api_url.username or parsed_base_url.username
        password = password or parsed_api_url.password or parsed_base_url.password
        if token:
            if not username:
                raise ValueError("Token authentication specified without username")
            if not self._token_checker.match(token):
                raise ValueError("Malformed Token String")
            self._api_headers["Authorization"] = f"Token {self._encode_auth_headers(username, token)}"
        elif password:
            if not username:
                raise ValueError("Password authentication specified without username")
            self._api_headers["Authorization"] = f"Basic {self._encode_auth_headers(username, password)}"
        elif username:
            raise ValueError("Username specified without authentication method")

    def _create_api_url(self, parts: List[str], query: Dict[str, str] = None) -> str:
        path = [self._api_path_prefix] + [quote(str(part).rstrip("/"), safe="") for part in parts]
        if query:
            path.append("")
            query = urlencode(query)
        return urlunsplit((self._api_scheme, self._api_netloc, "/".join(path), query, None))

    def _call(
        self, method: str, urlparts: List[str], urlquery: Dict[str, str] = None, body: Dict[str, Any] = None
    ) -> Dict[str, Any]:
        req_kwargs = {"headers": self._api_headers}
        if body:
            req_kwargs["json"] = body
        response = requests.request(method, self._create_api_url(urlparts, urlquery), **req_kwargs)
        self._check_api_response(response)
        return response.json()

    def upload_file(self, filepath: str) -> FileToken:
        with open(filepath, "rb") as file:
            response = requests.post(
                self._create_api_url(["uploads"]),
                files={"content": file},
                headers=self._api_headers,
            )
        self._check_api_response(response)
        return FileToken(response.json()["token"], filepath)

    def _create_data_url(self, rel_url: str) -> str:
        base_url = urlunsplit((self._url_scheme, self._url_netloc, self._url_path_prefix, None, None))
        return urljoin(base_url, rel_url)

    @classmethod
    def save_to_config(cls, config_name: str, **constructor_args) -> None:
        if not config_name.isalnum():
            raise ValueError("config_name must be alphanumeric")
        pathdir = user_data_dir(appname="pyszuru", appauthor=False, roaming=False)
        os.makedirs(pathdir, exist_ok=True)
        path = os.path.join(pathdir, f"{config_name}.json")
        with open(path, "w") as f:
            json.dump(constructor_args, f)

    @classmethod
    def load_from_config(cls, config_name: str) -> API:
        if not config_name.isalnum():
            raise ValueError("config_name must be alphanumeric")
        pathdir = user_data_dir(appname="pyszuru", appauthor=False, roaming=False)
        path = os.path.join(pathdir, f"{config_name}.json")
        with open(path, "r") as f:
            constructor_args = json.load(f)
        return cls(**constructor_args)



import setuptools
import subprocess


def get_git_version():
    return subprocess.run(
        ["git", "describe", "--abbrev=0"],
        capture_output=True,
        text=True,
        check=True,
    ).stdout.strip()


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyszuru",
    version=get_git_version(),
    author="Shyam Sunder",
    author_email="sgsunder1@gmail.com",
    description="Python interface for szurubooru",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sgsunder/python-szurubooru",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)



# flake8: noqa
from .api import API, FileToken, SzurubooruHTTPError
from .resource import Resource, ResourceNotSynchronized
from .tag import Tag
from .post import Post
from .search import search_tag, search_post, search_by_image, SearchResult
